#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

//#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>
#include <math.h>

//Mode3

//---------------------------------------------------------------------------
#define STATE_INIT 				1		//is
#define STATE_RESET				2
#define STATE_GAME				3

#define PI						3.1415926

#define MODE3_COLOR_WHITE		RGB(31,31,31)
#define MODE3_COLOR_BLOCK		RGB( 0, 0, 0)

//---------------------------------------------------------------------------
typedef struct {
	f32 x;
	f32 y;
} ST_POINT;

typedef struct {
	f32 x[4];
	f32 y[4];
} ST_CUBE;

typedef struct {
	f32 m11, m12, m13;
	f32 m21, m22, m23;
	f32 m31, m32, m33;
} ST_MATRIX;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_CUBE Cube;
u16     Key;
u8      State;						//is󋵃tO
//---------------------------------------------------------------------------
void StateInit(); 
void StateReset();
void StateGame(); 

void InitWS();
void InitKey();
void IRQUserHandler();

void RotateCube(ST_CUBE* p, s16 centerX, s16 centerY, s16 vect);
void DrawCube  (ST_CUBE* p);

f32  GetRadian(f32 deg);

void DebugDispMatrix(ST_MATRIX m);

ST_MATRIX MatrixGetRotate(f32 rad);
ST_MATRIX MatrixGetTranslate(f32 tx, f32 ty);
ST_MATRIX MatrixMulti(ST_MATRIX a, ST_MATRIX b);

ST_POINT  PointTransForm(ST_POINT p, ST_MATRIX m);

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RESET: StateReset(); break;
			case STATE_GAME:  StateGame();  break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitKey();

	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();
	Key = KEY_NONE;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;
	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void ScanKey()
{
	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return;
	}

	Key = KeyInput();
}
//---------------------------------------------------------------------------
void StateReset()
{
	//lp`
	Cube.x[0] =  60; Cube.y[0] =  50;
	Cube.x[1] = 180; Cube.y[1] =  50;
	Cube.x[2] = 180; Cube.y[2] = 110;
	Cube.x[3] =  60; Cube.y[3] = 110;

//	WaitForVsync();
	DrawCube(&Cube);

	State = STATE_GAME;
}
//---------------------------------------------------------------------------
void StateGame()
{
	ScanKey();

	if( (Key & KEY_LEFT) || (Key & KEY_RIGHT) )
	{
		s16 vectDeg = 0;

		if(Key & KEY_LEFT)  vectDeg  -= 10;
		if(Key & KEY_RIGHT) vectDeg  += 10;

		vectDeg = (vectDeg + 360) % 360;

		RotateCube(&Cube, SCREEN_CX/2, SCREEN_CY/2, vectDeg);

//		WaitForVsync();
		DrawCube(&Cube);
	}
}
//---------------------------------------------------------------------------
void RotateCube(ST_CUBE* p, s16 centerX, s16 centerY, s16 vect)
{
	f32       r;
	ST_MATRIX mm;
	ST_MATRIX mr;
	ST_MATRIX mcp;
	ST_MATRIX mcm;

	r   = GetRadian(vect);
	mr  = MatrixGetRotate(r);
	mcp = MatrixGetTranslate( 1*centerX,  1*centerY);
	mcm = MatrixGetTranslate(-1*centerX, -1*centerY);

	//{X ~ ] ~ |X ̏Ԃō܂
	mm  = MatrixMulti(mcp, mr);
	mm  = MatrixMulti(mm, mcm);

	//]̌vZ܂
	ST_POINT pt;
	s16      i;

	for(i=0; i<4; i++)
	{
		pt.x = p->x[i];
		pt.y = p->y[i];

		pt = PointTransForm(pt, mm);

		p->x[i] = pt.x;
		p->y[i] = pt.y;
	}
}
//---------------------------------------------------------------------------
void DrawCube(ST_CUBE* p)
{
	Mode3DrawCls(MODE3_COLOR_BLOCK);

	Mode3DrawLine(p->x[0], p->y[0], p->x[1], p->y[1], MODE3_COLOR_WHITE);
	Mode3DrawLine(p->x[1], p->y[1], p->x[2], p->y[2], MODE3_COLOR_WHITE);
	Mode3DrawLine(p->x[2], p->y[2], p->x[3], p->y[3], MODE3_COLOR_WHITE);
	Mode3DrawLine(p->x[3], p->y[3], p->x[0], p->y[0], MODE3_COLOR_WHITE);
}
//---------------------------------------------------------------------------
//ړsԂ܂
ST_MATRIX MatrixGetTranslate(f32 tx, f32 ty)
{
	ST_MATRIX ret;

	ret.m11 = 1; ret.m12 = 0; ret.m13 = tx;
	ret.m21 = 0; ret.m22 = 1; ret.m23 = ty;
	ret.m31 = 0; ret.m32 = 0; ret.m33 = 1;

	return ret;
}
//---------------------------------------------------------------------------
//]sԂ܂
ST_MATRIX MatrixGetRotate(f32 rad)
{
	ST_MATRIX ret;

	ret.m11 = cos(rad); ret.m12 = -sin(rad); ret.m13 = 0;
	ret.m21 = sin(rad); ret.m22 =  cos(rad); ret.m23 = 0;
	ret.m31 = 0;        ret.m32 =  0;        ret.m33 = 1;

	return ret;
}
//---------------------------------------------------------------------------
//s܂
ST_MATRIX MatrixMulti(ST_MATRIX a, ST_MATRIX b)
{
	ST_MATRIX ret;

	ret.m11 = (a.m11 * b.m11) + (a.m12 * b.m21) + (a.m13 * b.m31);
	ret.m12 = (a.m11 * b.m12) + (a.m12 * b.m22) + (a.m13 * b.m32);
	ret.m13 = (a.m11 * b.m13) + (a.m12 * b.m23) + (a.m13 * b.m33);

	ret.m21 = (a.m21 * b.m11) + (a.m22 * b.m21) + (a.m23 * b.m31);
	ret.m22 = (a.m21 * b.m12) + (a.m22 * b.m22) + (a.m23 * b.m32);
	ret.m23 = (a.m21 * b.m13) + (a.m22 * b.m23) + (a.m23 * b.m33);

	ret.m31 = (a.m31 * b.m11) + (a.m32 * b.m21) + (a.m33 * b.m31);
	ret.m32 = (a.m31 * b.m12) + (a.m32 * b.m22) + (a.m33 * b.m32);
	ret.m33 = (a.m31 * b.m13) + (a.m32 * b.m23) + (a.m33 * b.m33);

	return ret;
}
//---------------------------------------------------------------------------
//Wϊ܂
ST_POINT PointTransForm(ST_POINT p, ST_MATRIX m)
{
	ST_POINT ret;

	ret.x = (m.m11 * p.x) + (m.m12 * p.y) + m.m13;
	ret.y = (m.m21 * p.x) + (m.m22 * p.y) + m.m23;

	return ret;
};
//---------------------------------------------------------------------------
f32 GetRadian(f32 deg)
{
	return deg * (f32)PI / 180.0;
}
//---------------------------------------------------------------------------
void DebugDispMatrix(ST_MATRIX m)
{
	char buf[64];

	dprintf("----\n");
	bprintf(buf, "%f, %f, %f", m.m11, m.m12, m.m13); dprintf("%s\n", buf);
	bprintf(buf, "%f, %f, %f", m.m21, m.m22, m.m23); dprintf("%s\n", buf);
	bprintf(buf, "%f, %f, %f", m.m31, m.m32, m.m33); dprintf("%s\n", buf);
	dprintf("----\n\n");

}
